import os
import tempfile

from celery.utils.log import get_task_logger
from django.conf import settings
from rgd.models import ChecksumFile
from rgd.utility import get_or_create_no_commit
from rgd_3d.models import PointCloud, PointCloudMeta

logger = get_task_logger(__name__)


def _file_conversion_helper(source, output_field, method, prefix='', extension='', **kwargs):
    workdir = getattr(settings, 'GEODATA_WORKDIR', None)
    with tempfile.TemporaryDirectory(dir=workdir) as tmpdir:
        with source.yield_local_path() as file_path:
            output_path = os.path.join(tmpdir, prefix + os.path.basename(source.name) + extension)
            method(str(file_path), str(output_path), **kwargs)
        with open(output_path, 'rb') as f:
            output_field.save(os.path.basename(output_path), f)


def _save_pyvista(mesh, output_path):
    import pyvista as pv

    points = pv.PolyData(mesh.points)
    points.point_arrays.update(mesh.point_arrays)
    points.save(output_path)


def _use_pyntcloud_pyvista(input_path, output_path):
    from pyntcloud import PyntCloud

    cloud = PyntCloud.from_file(input_path)
    mesh = cloud.to_instance('pyvista', mesh=False)
    _save_pyvista(mesh, output_path)


def _use_pyvista(input_path, output_path):
    import pyvista as pv

    mesh = pv.read(input_path)
    _save_pyvista(mesh, output_path)


def _get_readers():
    import pyvista as pv

    methods = {
        'las': _use_pyntcloud_pyvista,
        'npy': _use_pyntcloud_pyvista,
        'npz': _use_pyntcloud_pyvista,
    }
    methods.update({k[1::]: _use_pyvista for k in pv.utilities.fileio.READERS.keys()})
    return methods


def read_point_cloud_file(pc_file):
    """Read a PointCloud object and create a new PointCloudMeta."""
    if not isinstance(pc_file, PointCloud):
        pc_file = PointCloud.objects.get(pk=pc_file)
    pc_entry, _ = get_or_create_no_commit(PointCloudMeta, source=pc_file)
    pc_entry.name = pc_file.file.name
    # Parse the point cloud file format and convert
    ext = pc_file.file.name.split('.')[-1].strip().lower()
    try:
        method = _get_readers()[ext]
    except KeyError:
        raise ValueError(f'Extension `{ext}` unsupported for point cloud conversion.')
    try:
        pc_entry.vtp_data
    except ChecksumFile.DoesNotExist:
        pc_entry.vtp_data = ChecksumFile()
    if not pc_entry.vtp_data.collection:
        pc_entry.vtp_data.collection = pc_file.file.collection
    _file_conversion_helper(pc_file.file, pc_entry.vtp_data.file, method, extension='.vtp')
    # Save the record
    pc_entry.save()



import inspect

import factory
from pytest_factoryboy import register
from rgd_testing_utils import factories
from rgd_testing_utils.data_fixtures import *  # noqa

for _, fac in inspect.getmembers(factories):
    if inspect.isclass(fac) and issubclass(fac, factory.Factory):
        register(fac)



import logging

from rgd.management.commands._data_helper import SynchronousTasksCommand, _get_or_create_file_model
from rgd_3d import models

logger = logging.getLogger(__name__)


SUCCESS_MSG = 'Finished loading all demo data.'

# Names of files in the datastore
POINT_CLOUD_FILES = [
    'topo.vtk',
]


def load_point_cloud_files(pc_files):
    ids = []
    for f in pc_files:
        entry = _get_or_create_file_model(models.PointCloud, f)
        ids.append(entry.pk)
    return ids


class Command(SynchronousTasksCommand):
    help = 'Populate database with demo data.'

    def handle(self, *args, **options):

        self.set_synchronous()
        # Run the command
        load_point_cloud_files(POINT_CLOUD_FILES)

        self.stdout.write(self.style.SUCCESS(SUCCESS_MSG))
        self.reset_celery()



from django.urls import path
from rgd_3d import models, rest, views

urlpatterns = [
    # Pages
    path(
        'rgd_3d/point_cloud/<int:pk>/',
        views.PointCloudMetaDetailView.as_view(),
        name=models.PointCloudMeta.detail_view_name,
    ),
    #############
    # Other
    path(
        'api/rgd_3d/point_cloud/<int:pk>',
        rest.get.GetPointCloudMeta.as_view(),
        name='point-cloud-entry',
    ),
    path(
        'api/rgd_3d/point_cloud/<int:pk>/base64',
        rest.get.GetPointCloudMetaData.as_view(),
        name='point-cloud-entry-data',
    ),
]



from celery import shared_task
from rgd.tasks import helpers


@shared_task(time_limit=86400)
def task_read_point_cloud_file(pc_file_pk):
    from rgd_3d.models import PointCloud

    from .etl import read_point_cloud_file

    pc_file = PointCloud.objects.get(pk=pc_file_pk)
    helpers._run_with_failure_reason(pc_file, read_point_cloud_file, pc_file_pk)



from django.apps import apps
from rgd_testing_utils.helpers import check_model_permissions


def test_check_permissions_path_rgd_3d():
    for model in apps.get_app_config('rgd_3d').get_models():
        check_model_permissions(model)



from rest_framework.generics import RetrieveAPIView
from rgd.rest.get import _PermissionMixin
from rgd_3d import models, serializers


class GetPointCloudMeta(RetrieveAPIView, _PermissionMixin):
    serializer_class = serializers.PointCloudMetaSerializer
    lookup_field = 'pk'
    queryset = models.PointCloudMeta.objects.all()


class GetPointCloudMetaData(RetrieveAPIView, _PermissionMixin):
    serializer_class = serializers.PointCloudMetaDataSerializer
    lookup_field = 'pk'
    queryset = models.PointCloudMeta.objects.all()



from django.contrib.gis.db import models
from django.contrib.postgres import fields
from django_extensions.db.models import TimeStampedModel
from rgd.models import ChecksumFile, SpatialEntry
from rgd.models.mixins import DetailViewMixin, PermissionPathMixin, TaskEventMixin
from rgd_3d.tasks import jobs


class PointCloud(TimeStampedModel, TaskEventMixin, PermissionPathMixin):
    """Container for point cloud file."""

    task_funcs = (jobs.task_read_point_cloud_file,)
    file = models.ForeignKey(ChecksumFile, on_delete=models.CASCADE, related_name='+')

    def data_link(self):
        return self.file.data_link()

    data_link.allow_tags = True

    permissions_paths = [('file', ChecksumFile)]


class PointCloudMeta(TimeStampedModel, PermissionPathMixin, DetailViewMixin):
    """Container for converted point cloud data.

    The data here must be stored in VTP format. This can be manually uploaded
    or created automatically via a PointCloud upload.
    """

    name = models.CharField(max_length=1000, blank=True)
    description = models.TextField(null=True, blank=True)

    # Can be null if not generated from uploaded file
    source = models.OneToOneField(PointCloud, null=True, blank=True, on_delete=models.CASCADE)

    # A place to store converted file - must be in VTP format
    vtp_data = models.ForeignKey(ChecksumFile, on_delete=models.DO_NOTHING, related_name='+')

    def data_link(self):
        return self.vtp_data.data_link()

    data_link.allow_tags = True

    permissions_paths = [('source', PointCloud), ('vtp_data', ChecksumFile)]
    detail_view_name = 'point-cloud-entry-detail'


class PointCloudSpatial(TimeStampedModel, SpatialEntry, PermissionPathMixin):
    """Optionally register a PointCloudMeta as a SpatialEntry."""

    source = models.OneToOneField(PointCloud, on_delete=models.CASCADE)

    crs = models.TextField(help_text='PROJ string', blank=True, null=True)  # PROJ String
    # Origin point to map the 0,0,0 point of the point cloud
    origin = fields.ArrayField(models.FloatField(), size=3, blank=True, null=True)

    @property
    def name(self):
        return self.source.file.name

    permissions_paths = [('source', PointCloud)]
    detail_view_name = 'point-cloud-entry-detail'
    detail_view_pk = 'source__pointcloudmeta__pk'



import pytest
from rgd.datastore import datastore
from rgd_3d import models
from rgd_3d.tasks.etl import read_point_cloud_file

from . import factories


@pytest.mark.parametrize(
    'sample_file',
    [
        'topo.vtk',
    ],
)
@pytest.mark.django_db(transaction=True)
def test_point_cloud_etl(sample_file):
    pc_file = factories.PointCloudFactory(
        file__file__filename=sample_file,
        file__file__from_path=datastore.fetch(sample_file),
    )
    entry = models.PointCloudMeta.objects.filter(source=pc_file).first()
    assert entry.vtp_data is not None
    # Testing that we can repopulate a point cloud entry
    read_point_cloud_file(pc_file.id)



import base64

from rest_framework import serializers
from rgd import utility
from rgd.serializers import ChecksumFileSerializer
from rgd_3d import models


class PointCloudSerializer(serializers.ModelSerializer):
    file = ChecksumFileSerializer()

    class Meta:
        model = models.PointCloud
        fields = '__all__'


class PointCloudMetaSerializer(serializers.ModelSerializer):
    source = PointCloudSerializer()
    vtp_data = ChecksumFileSerializer(required=False)

    def to_representation(self, value):
        ret = super().to_representation(value)
        return ret

    class Meta:
        model = models.PointCloudMeta
        fields = '__all__'


class PointCloudMetaDataSerializer(PointCloudMetaSerializer):
    def to_representation(self, value):
        ret = super().to_representation(value)
        # Extract data as base64
        with value.vtp_data.yield_local_path() as path:
            with open(path, 'rb') as data:
                data_content = data.read()
                base64_content = base64.b64encode(data_content)
                base64_content = base64_content.decode().replace('\n', '')
        ret['vtp_data'] = base64_content
        return ret


utility.make_serializers(globals(), models)



from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.admin.mixins import (
    MODIFIABLE_FILTERS,
    TASK_EVENT_FILTERS,
    TASK_EVENT_READONLY,
    _FileGetNameMixin,
    reprocess,
)
from rgd_3d.models import PointCloud, PointCloudMeta, PointCloudSpatial


@admin.register(PointCloud)
class PointCloudAdmin(OSMGeoAdmin, _FileGetNameMixin):
    list_display = (
        'pk',
        'get_name',
        'status',
        'modified',
        'created',
        'data_link',
    )
    readonly_fields = (
        'modified',
        'created',
    ) + TASK_EVENT_READONLY
    actions = (reprocess,)
    list_filter = MODIFIABLE_FILTERS + TASK_EVENT_FILTERS


@admin.register(PointCloudSpatial)
class PointCloudSpatialAdmin(OSMGeoAdmin):
    list_display = (
        'pk',
        'modified',
        'created',
        'crs',
    )
    readonly_fields = (
        'modified',
        'created',
    )
    modifiable = False  # To still show the footprint and outline


@admin.register(PointCloudMeta)
class PointCloudMetaAdmin(OSMGeoAdmin):
    list_display = (
        'pk',
        'modified',
        'created',
        'data_link',
    )
    readonly_fields = (
        'source',
        'modified',
        'created',
    )
    list_filter = MODIFIABLE_FILTERS



from django.apps import AppConfig


class RGD3DConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'rgd_3d'

    def ready(self):
        import rgd_3d.signals  # noqa: F401



from django.db import transaction
from django.db.models.signals import post_save
from django.dispatch import receiver
from rgd.utility import skip_signal
from rgd_3d import models


@receiver(post_save, sender=models.PointCloud)
@skip_signal()
def _post_save_point_cloud_file(sender, instance, *args, **kwargs):
    transaction.on_commit(lambda: instance._post_save_event_task(*args, **kwargs))
